#include <stdbool.h>
#include <stdint.h>

#include "example-main.h"
#include "libti2cit.h"

#include "inc/hw_i2c.h"
#include "inc/hw_ints.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/gpio.h"
#include "driverlib/i2c.h"
#include "driverlib/rom.h"

typedef struct example_isr_st_ {
	libti2cit_int_st ti2cit;
	uint32_t scan_addr;
	uint32_t scan_found;
	uint32_t talk_count;
	uint32_t sysclock;
} example_isr_st;

example_isr_st i2c7;

void memzero(void * dst, uint32_t len)
{
	uint8_t * p = dst;
	while (len--) *p++ = 0;
}

void main_isrnofifo(uint32_t sysclock)
{
	// initialize to 0
	memzero(&i2c7, sizeof(i2c7));

	i2c7.ti2cit.base = I2C7_BASE;
	i2c7.sysclock = sysclock;	// i2c7.sysclock is not used currently

	// initialize I2C7 slave and turn on loopback mode
	ROM_I2CSlaveInit(i2c7.ti2cit.base, 0x7f);

	libti2cit_s_int_clear(&i2c7.ti2cit);

	ROM_IntMasterEnable();
	ROM_IntEnable(INT_I2C7);
	ROM_I2CSlaveIntEnableEx(i2c7.ti2cit.base, I2C_SIMR_STARTIM | I2C_SIMR_STOPIM | I2C_SIMR_DATAIM);

	UARTsend("waiting for interrupts\r\n");
	for (;;);	// now just wait for slave interrupts
}

static void i2cInt_simr_dump(uint32_t status)
{
	static char buf[] = "status=0000";
	u16tohex(&buf[7], status);
	buf[11] = 0;
	UARTsend(buf);

	if (status & I2C_SIMR_DATAIM) UARTsend(" RIS");
	if (status & I2C_SIMR_STARTIM) UARTsend(" start");
	if (status & I2C_SIMR_STOPIM) UARTsend(" stop");
	if (status & I2C_SIMR_DMARXIM) UARTsend(" dmaRX");
	if (status & I2C_SIMR_DMATXIM) UARTsend(" dmaTX");
	if (status & I2C_SIMR_TXIM) UARTsend(" tx");
	if (status & I2C_SIMR_RXIM) UARTsend(" rx");
	if (status & I2C_SIMR_TXFEIM) UARTsend(" txFE");
	if (status & I2C_SIMR_RXFFIM) UARTsend(" rxFF");

	UARTsend("\r\n");
}

uint8_t slave_buf[4];
uint32_t slave_buf_pos = 0;
uint8_t master_buf[4];
uint32_t master_buf_pos = 0;
void i2c7Int_isrnofifo()
{
	uint32_t status = libti2cit_s_isr_isr(&i2c7.ti2cit);

	// if the status is only START or only STOP or only START|STOP, there is nothing to do but UARTsend() and return
	if (status == LIBTI2CIT_ISR_S_START) {
		UARTsend("i2c7 start\r\n");
		master_buf_pos = 0;
		slave_buf_pos = 0;
		return;
	}
	if (status == LIBTI2CIT_ISR_S_STOP) {
		UARTsend("i2c7 stop\r\n");
		master_buf_pos = 0;
		slave_buf_pos = 0;
		return;
	}
	if (status == (LIBTI2CIT_ISR_S_START | LIBTI2CIT_ISR_S_STOP)) return;

	// the status has something else... this code never seems to run, so do lots and lots of UARTsend() here
	if (status & LIBTI2CIT_ISR_S_START) {
		UARTsend("i2c7 S_START ");
	}
	if (status & LIBTI2CIT_ISR_S_STOP) {
		UARTsend("i2c7 S_STOP ");
	}
	if (status & I2C_SCSR_QCMDST) {
		master_buf_pos = 0;
		slave_buf_pos = 0;

		if (status & I2C_SCSR_QCMDRW) {
			// master is requesting to read bytes
			UARTsend("s_QUICK read\r\n");
		} else {
			UARTsend("s_QUICK\r\n");
		}

		// is it possible to NAK a QCMD?
		uint32_t sack = (HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) | I2C_SACKCTL_ACKOEN) & (~I2C_SACKCTL_ACKOVAL);
		HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) = sack | I2C_SACKCTL_ACKOVAL;	// send NACK
		HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) = (sack | I2C_SACKCTL_ACKOVAL) & (~I2C_SACKCTL_ACKOEN);
	} else if (status & I2C_SCSR_QCMDRW) {
		UARTsend("s_QUICK read outside s_QUICK ");
	}

	if (status & I2C_SCSR_RREQ) {
		if (status & I2C_SCSR_FBR) {
			// start of a buf from master
			master_buf_pos = 0;
			slave_buf_pos = 0;
			UARTsend("s_1st ");
		}
		// read the byte in the buf
		uint32_t sack = (HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) | I2C_SACKCTL_ACKOEN) & (~I2C_SACKCTL_ACKOVAL);
		if (master_buf_pos >= sizeof(master_buf)) {
			(void) HWREG(i2c7.ti2cit.base + I2C_O_SDR);
			HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) = sack | I2C_SACKCTL_ACKOVAL;	// send NACK
			UARTsend("s_NACK\r\n");
		} else {
			master_buf[master_buf_pos++] = HWREG(i2c7.ti2cit.base + I2C_O_SDR);
			HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) = sack;	// send ACK
			HWREG(i2c7.ti2cit.base + I2C_O_SACKCTL) = sack & (~I2C_SACKCTL_ACKOEN);
			UARTsend("s_ACK\r\n");
		}
		// signal the non-isr code to look at the buf
	}
	if (status & I2C_SCSR_TREQ) {
		UARTsend("s_TREQ ");
		i2cInt_simr_dump(HWREG(i2c7.ti2cit.base + I2C_O_SIMR));

		/* NOTE: it is a bad idea to do anything more complicated than copying data out of a buffer here
		 * the i2c master can time out if you delay long enough before responding to a _TREQ
		 */
		HWREG(i2c7.ti2cit.base + I2C_O_SDR) = (slave_buf_pos >= sizeof(slave_buf)) ? 0xff : (/*slave_buf[*/slave_buf_pos++);
		// TODO: master can NACK
	}
	if (status & LIBTI2CIT_ISR_UNEXPECTED) {
		UARTsend("isr_s unexpected\r\n");
	}
}
